--[[                         Impulse
             Project started on 26. Sep 2009
                dev-Version of 26. Nov 2009

Copyright 2009 Daniel Sturm <skasski@gmx.at> ]]

l_g = love.graphics
setColor = l_g.setColor
newColor = l_g.newColor

love.filesystem.require("camera.lua")
	worldCam = camera.new(10, 10, 0.5, 0.5, 0.5, 0.5)
	zoomCam = camera.new(10, 10, 0.5, 0.5, 0.5, 0.5)
	guiCam = camera.stretchToResolution(l_g.getWindowWidth(),
										l_g.getWindowHeight())

love.filesystem.require("input.lua")
love.filesystem.require("controls.lua")

gui = {}
gui.bar = l_g.newImage("gui/bar.png")
circle = l_g.newImage("circle.png")

-- love.audio.setChannels(2)
-- sound = {}
-- sound.absorb = love.audio.newSound("audio/absorb.wav")
-- sound.absorbed = love.audio.newSound("audio/absorbed.wav")


-- function read_multiple(path)
	-- multiple = {}
	-- local i = 1
	-- local currline = 1
	-- s = love.filesystem.read(path)
	-- while true do
		-- nextline = string.find(s, "\n", currline)
		-- if nextline == nil then
			-- multiple[i] = string.sub(s, currline)
			-- break
		-- end
		-- multiple[i] = string.sub(s, currline, nextline-2)
		-- currline = nextline+1
		-- i = i +1
	-- end
-- end

function kill(i)
	if entities[i].breed == "player" then
		player.m = 0
		player.r = 0
	end
	local x, y = getBox(entities[i].x, entities[i].y)
	if boxes[x][y][i] then
		boxes[x][y][i] = nil
	end
	entities[i] = nil
end

function radius(mass)
	--return (mass/math.pi/(4/3))^(1/3)
	return (mass/math.pi)^0.5
end

function mass(radius)
	--return (4/3)*(radius^3)*math.pi
	return (radius^2)*math.pi
end		

function origin()
	local x = player.x
	local y = player.y
	local width = world_width*1.1
	local height = world_height*1.1
	
	-- set camera origin
	-- if view is bigger then the world size, center middle
	if l_g:getWidth()/2 >= width then
		x = 0
	-- if playerposition + view is bigger than the world size, move origin
	elseif x - l_g:getWidth()/2 < -width then
		x = -width+l_g:getWidth()/2
	-- if playerposition + view is bigger than the world size, move origin
	elseif x + l_g:getWidth()/2 > width then
		x = width-l_g:getWidth()/2
	end
	-- if view is bigger then the world size, center middle
	if l_g:getHeight()/2 >= height then
		y = 0
	-- if playerposition + view is bigger than the world size, move origin
	elseif y - l_g:getHeight()/2 < -height then
		y = -height+l_g:getHeight()/2
	-- if playerposition + view is bigger than the world size, move origin
	elseif y + l_g:getHeight()/2 > height then
		y = height-l_g:getHeight()/2
	end
	getCamera():setOrigin(x, y)
end

function absorb(bigger, smaller)
	-- calculate the distance between bigger and smaller
	local d = ((bigger.x-smaller.x)^2+(bigger.y-smaller.y)^2)^0.5
	local r = bigger.r+smaller.r
	if d < r then -- breeds absorb each other
		-- check whether "bigger" is actually the bigger body
		if bigger.m < smaller.m then -- if not then flip!
			bigger, smaller = smaller, bigger
		end
		-- gravitators always absorb default breeds!
		if smaller.breed == "gravitator" and bigger.breed == "default" then
			bigger, smaller = smaller, bigger
		end
		
		--  now start the progress - absorb - yummy!
		-- decrease the radius of smaller 
		smaller.r = d-bigger.r
		-- calculate the absorbed matter based on the new radius
		local matter = smaller.m - math.max(0, mass(math.max(0, smaller.r)))
		-- make absorbed mass affect the speed, if not a gravitator absorbs
		if bigger.breed ~= "gravitator" then
			bigger.sx = (matter*smaller.sx + bigger.m*bigger.sx)/(matter+bigger.m)
			bigger.sy = (matter*smaller.sy + bigger.m*bigger.sy)/(matter+bigger.m)
		end
		-- add mass
		bigger.m = bigger.m + matter
		-- zoom out, if entity of player
		if bigger.i == player.i then
			-- love.audio.play(sound.absorb)
			if math.min(l_g:getWidth(), l_g:getHeight()) < radius(bigger.m)*12
			 and ((world_width*1.1)*2 > l_g:getWidth()* (radius(bigger.m)/bigger.r)
			 or  (world_height*1.1)*2 > l_g:getHeight()* (radius(bigger.m)/bigger.r)) then
				scale = getCamera():getScaleFactor()*math.max(1/(radius(bigger.m)/bigger.r),
															  (math.min(l_g:getWidth(), l_g:getHeight())/(radius(bigger.m)*12)))
			end
		elseif smaller.i == player.i then
			-- love.audio.play(sound.absorbed)
		end
		-- calculate new radius
		bigger.r = radius(bigger.m)
		-- update bigger
		entities[bigger.i] = bigger
		if d < bigger.r then -- remove smaller, if fully absorbed
			kill(smaller.i)
			if bigger.i == player.i then
				multi_counter = multi_counter +1
				multi_timer = 3
			end
		else -- otherwise calculate its mass and update smaller too
			smaller.m = mass(smaller.r)
			entities[smaller.i] = smaller
		end
		
	end
end

mouse = l_g.newImage("mouse.png")
love.mouse.setVisible(false)

function create_entity(breed, m, x, y, sx, sy)
	sx = sx or 0
	sy = sy or 0
	local r = radius(m)
	if x+r > world_width then
		x = world_width-r
	elseif x-r < -world_width then
		x = -world_width+r
	end
	if y+r > world_height then
		y = world_height-r
	elseif y-r < -world_height then
		y = -world_height+r
	end
	table.insert(entities, {breed = breed, m = m, r=r,
	                        x = x, y = y,
							sx = sx, sy = sy})
end

function createboxes(mass)
	box_amount_width  = math.floor(world_width / radius(mass)/2)
	if box_amount_width > world.entities^0.6 then
		box_amount_width = math.floor(world.entities^0.6)
	end
	box_amount_height = math.floor(world_height / radius(mass)/2)
	if box_amount_height > world.entities^0.6 then
		box_amount_height = math.floor(world.entities^0.6)
	end
	
	world_boxwidth = world_width / box_amount_width
	world_boxheight = world_height / box_amount_height
	
	boxes = {}
	for x=1, box_amount_width*2 +2 do
		boxes[x] = {}
		for y=1, box_amount_height*2 +2 do
			boxes[x][y] = {}
		end
	end
end

function toBox(i, x, y)
	local x = math.ceil((x+world_width+world_boxwidth)/world_boxwidth)
	local y = math.ceil((y+world_height+world_boxheight)/world_boxheight)
	boxes[x][y][i] = i
end
function getBox(x, y)
	local x = math.ceil((x+world_width+world_boxwidth)/world_boxwidth)
	local y = math.ceil((y+world_height+world_boxheight)/world_boxheight)
	return x, y
end

map = "beginning"
-- read_multiple("multiple.txt")
-- MAIN FUNCTIONS
function load()
	cpu = {}
	cpu.entities = 0
	cpu.drawentities = 0
	cpu.DEsetup = 0
	cpu.DEmisc = 0
	cpu.absorb = 0
	cpu.createboxes = 0
	cpu.gui = 0
	
	game = {}
	game.time = 0
	game.speed = 1
	
	multi_timer = 0
	multi_counter = 0
	
	information = ""
	entities = {}
	
	-- load up a world
	setCamera(worldCam)
	world = {}
	world.entities = 0
	world.biggest_amount = 0
	world.biggest_radius = 0
	love.filesystem.include("worlds/"..map..".lua")
	
	for i, v in pairs(entities) do
		world.entities = world.entities +1
		if v.m > world.biggest_amount then
			world.biggest_amount = v.m
			world.biggest_radius = v.r
			world.biggest_entity = i
		end
	end
	
	createboxes(world.biggest_amount)
	
	player = {}
	player.x = 0
	player.y = 0
	
	
	local width = l_g:getWidth()*getCamera():getScaleFactor()
	local height = l_g:getHeight()*getCamera():getScaleFactor()
	getCamera():setScaleFactor(math.min(width  / ((world_width*1.2)*2),
		                              height / ((world_height*1.2)*2) )*.01 )
	scale = math.min(width  / ((world_width*1.2)),
		            height / ((world_height*1.2)) )
	
	world.matter = 0
	
	
	-- set font
	setCamera(guiCam)
	l_g.setFont(l_g.newFont(love.default_font, l_g.getHeight()*0.05))
end

function draw()
-- draw world
	l_g.setColorMode(love.color_modulate)
	setCamera(guiCam)
	-- draw square
	l_g.setLine(4) 
	setCamera(worldCam)
	world.draw()
	
	-- draw boxes
	-- for i, v in pairs(boxes) do
		-- for j, w in pairs(boxes[i]) do
			-- l_g.rectangle(1, -world_width-world_boxwidth+(i-1)*world_boxwidth, -world_height-world_boxheight+(j-1)*world_boxheight,
			                -- world_boxwidth, world_boxheight)
		-- end
	-- end
	
	cpu.drawentities = cpu.drawentities - love.timer.getTime()
	
	local x, y = getCamera():getOrigin()
	local width = l_g.getWidth()/2
	local height = l_g.getHeight()/2
	for i, v in pairs(entities) do
		if v.x-v.r < x + width
		 and v.x+v.r > x - width
		 and v.y-v.r < y + height
		 and v.y+v.r > y - height
		 then
			breeds[v.breed].draw(v, i)
		end
	end
	cpu.drawentities = cpu.drawentities + love.timer.getTime()
	
	-- Draw loupe
	if love.mouse.isDown(love.mouse_right) then
		local zoom = 3
		local loupesize = 0.25
		
		setCamera(worldCam)
		local scale = getCamera():getScaleFactor()
		local worldX, worldY = love.mouse.getPosition()
		local width = l_g.getWidth()/2/(zoom/loupesize)
		local height = l_g.getHeight()/2/(zoom/loupesize)
		
		setCamera(guiCam)
		local guiX, guiY = love.mouse.getPosition()
		local guiX = guiX / l_g.getWidth()
		local guiY = guiY / l_g.getHeight()
		
		setCamera(zoomCam)
		getCamera():setOrigin(worldX, worldY)
		getCamera():setScreenOrigin(guiX, guiY)
		getCamera():setScaleFactor(scale*zoom, scale*zoom)
		l_g.setBlendMode(love.blend_normal)
		setColor(newColor(0, 0, 0))
		l_g.rectangle(love.draw_fill, worldX-width, worldY-height, width*2, height*2)
		
		l_g.setScissor(worldX-width, worldY-height, width*2, height*2) 
		-- Execute the current worlds draw function
		world.draw()
		-- Draw entities if they are inside the screen
		for i, v in pairs(entities) do
			if v.x-v.r < worldX + width
			 and v.x+v.r > worldX - width
			 and v.y-v.r < worldY + height
			 and v.y+v.r > worldY - height
			 then
				breeds[v.breed].draw(v, i)
			end
		end
		l_g.setScissor()
		
		l_g.setBlendMode(love.blend_normal)
		setColor(newColor(128, 128, 128))
		l_g.rectangle(love.draw_line, worldX-width, worldY-height, width*2, height*2)
		setCamera(worldCam)
	end
	
	
	
-- draw gui
cpu.gui = cpu.gui - love.timer.getTime()
	-- draw mouse
	local x, y = love.mouse.getPosition()
	setCamera(guiCam)
	if not love.mouse.isDown(love.mouse_right) then
		local mx, my = love.mouse.getPosition()
		local degree = math.deg(math.atan2(player.y-y, player.x-x))
		setColor(newColor(255, 255, 255))
		l_g.draw(mouse, mx, my, degree+90)
	end
	
	
	-- startup text
	l_g.setBlendMode(love.blend_additive)
	if game.time < 5 and not love.keyboard.isDown(love.key_f1) then
		local text = "Become the biggest!"
		setColor(newColor(192, 255, 192))
		local x = l_g.getWidth()*0.5-(l_g.getFont():getWidth(text))/2
		local y = l_g.getHeight()*0.7
		l_g.draw(text, x, y)
	end
	if love.timer.getTime() < 10 and not love.keyboard.isDown(love.key_f1) then
		local text = "Hold down F1 for help!"
		setColor(newColor(255, 255, 255))
		local x = l_g.getWidth()*0.5-(l_g.getFont():getWidth(text))/2
		local y = l_g.getHeight()*0.4
		l_g.draw(text, x, y)
	end
	-- Show help when pressing F1
	if love.keyboard.isDown(love.key_f1) then
		local text = "Welcome to Impulse!"
		setColor(newColor(255, 255, 255))
		local x = l_g.getWidth()*0.5-(l_g.getFont():getWidth(text))/2
		local y = l_g.getHeight()*0.2
		l_g.draw(text, x, y)
		local text = "F1 for help\nF5 to load a new world\nBackspace to reset\nESC to exit"
		setColor(newColor(0, 255, 255))
		local x = l_g.getWidth()*0.3
		local y = l_g.getHeight()*0.4
		l_g.draw(text, x, y)
		local text = "Become the biggest!"
		setColor(newColor(192, 255, 192))
		local x = l_g.getWidth()*0.5-(l_g.getFont():getWidth(text))/2
		local y = l_g.getHeight()*0.7
		l_g.draw(text, x, y)
	end
	
	-- display win/loose text
	if game.status == "won" then
		setColor(newColor(50, 100, 255))
		local text = "You have won. Press F5 to replay."
		local x = l_g.getWidth()*0.5-(l_g.getFont():getWidth(text))/2
		local y = l_g.getHeight()*0.75
		l_g.draw(text, x, y)
	elseif game.status == "lost" then
		setColor(newColor(255, 150, 0))
		local text = "You have lost. Press F5 to retry."
		local x = l_g.getWidth()*0.5-(l_g.getFont():getWidth(text))/2
		local y = l_g.getHeight()*0.75
		l_g.draw(text, x, y)
	end
	
	
	-- gui
	l_g.setBlendMode(love.blend_normal)
	-- biggest entity
	setColor(newColor(255, 0, 0, 255))
	local gauge_height = l_g.getHeight() - l_g.getHeight()/radius(world.matter)*radius(game.goal)
	l_g.line(l_g.getWidth(), gauge_height,
					   l_g.getWidth()-24, gauge_height)
	-- player
	setColor(newColor(128, 128, 255, 192))
	gauge_height = l_g.getHeight() - l_g.getHeight()/radius(world.matter)*player.r
	l_g.quad(love.draw_fill,
	                   l_g.getWidth()-3, l_g.getHeight(),
					   l_g.getWidth()-3, gauge_height,
					   l_g.getWidth()-21, gauge_height,
					   l_g.getWidth()-21, l_g.getHeight()) 
	-- bar
	setColor(newColor(255, 255, 255, 128))
	l_g.draw(gui.bar, l_g.getWidth()-12, l_g.getHeight()/2, 0, 1, l_g.getHeight())
	
	
	-- speed
	setColor(newColor(255, 255, 128, 192))
	l_g.draw(game.speed.."x", 0, l_g.getFont():getHeight())
	
	
	-- multi-soaks
	-- if multi_counter > 1 then
		-- l_g.draw(multiple[math.min(multi_counter, #multiple)].."-soak ("..string.format("%.2f", multi_timer)..")", 200, 50)
	-- end
	
	
	-- amount of entities
	-- l_g.draw(world.entities, 100, 20)
	
	cpu.gui = cpu.gui + love.timer.getTime()
	
	
	-- draw cpu time
	-- setColor(newColor(0, 255, 0))
	-- local cpuTimes = "Time = "..game.time.."\n"
	-- for i, v in pairs (cpu) do
		-- cpuTimes = cpuTimes..i.."\n"..v.."\n"
	-- end
	-- l_g.draw(cpuTimes, 100, l_g.getFont():getHeight())
	
	--draw FPS
	l_g.draw(love.timer.getFPS().."fps", 0, l_g.getFont():getHeight()*2)
	-- draw matter
	-- l_g.draw("Matter: "..world.matter, 0, l_g.getFont():getHeight()*2)
	-- draw box amounts
	-- l_g.draw(box_amount_width.."/"..box_amount_height, 0, l_g.getFont():getHeight()*3)
	-- draw dt
	-- l_g.draw(duration, 0, l_g.getFont():getHeight()*3)
end

function update(dt)
	game.time = game.time+dt
	
	dt = math.min(dt, .2)
	
	setCamera(worldCam)
	if getCamera():getScaleFactor() > scale then
		-- zoom out
		newScaleFactor = getCamera():getScaleFactor() + (scale-getCamera():getScaleFactor())*(0.3-dt)
	else
		-- zoom in
		newScaleFactor = getCamera():getScaleFactor() + (scale-getCamera():getScaleFactor())*dt
	end
	getCamera():setScaleFactor(newScaleFactor, newScaleFactor)
	
	-- multiply dt with the game speed
	dt = dt*game.speed
	dt = math.min(dt, .2)
	duration = dt
	
	multi_timer = multi_timer -dt
	if not (multi_timer > 0) then
		multi_counter = 0
	end
	
	world.entities = 0
	world.matter = 0
	-- entities take action
	calculations = 0
	cpu.entities = cpu.entities-love.timer.getTime()
	world.biggest_amount = 0
	for i, v in pairs(entities) do
		v.i = i
		world.entities = world.entities + 1
		if v.m > world.biggest_amount then
			world.biggest_radius = v.r
			world.biggest_amount = v.m
			world.biggest_entity = i
		end
		
		-- move entity around
		v.x = v.x+v.sx*dt
		v.y = v.y+v.sy*dt
		
		-- breeds take action
		breeds[v.breed].task(v, i, dt)
	end
	
	world.update()
	
	cpu.createboxes = cpu.createboxes-love.timer.getTime()
	createboxes(world.biggest_amount)
	cpu.createboxes = cpu.createboxes+love.timer.getTime()
	
	for i, v in pairs(entities) do
		toBox(i, v.x, v.y)
	end
	cpu.entities = cpu.entities+love.timer.getTime()
	
	cpu.absorb = cpu.absorb-love.timer.getTime()
	for i, v in pairs(entities) do
		-- get the box the entity is part of
		local x, y = getBox(v.x, v.y)
		for i2, v2 in pairs(boxes[x][y]) do
			if v2 > i then
				absorb(v, entities[v2])
			end
		end
		for i2, v2 in pairs(boxes[x+1][y]) do
			absorb(v, entities[v2])
		end
		for i2, v2 in pairs(boxes[x-1][y+1]) do
			absorb(v, entities[v2])
		end
		for i2, v2 in pairs(boxes[x][y+1]) do
			absorb(v, entities[v2])
		end
		for i2, v2 in pairs(boxes[x+1][y+1]) do
			absorb(v, entities[v2])
		end
	end
	cpu.absorb = cpu.absorb+love.timer.getTime()
	
	-- position worldCamera
	origin()
end

camera.lateInit()