world.radius = 50

world_height = world.radius
world_width = world.radius
world.matter = (world.radius^2) * math.pi * 0.02

function world.draw()
	love.graphics.setColor(love.graphics.newColor(255, 255, 255, 128))
	love.graphics.draw(circle, 0, 0, 0, world.radius/500)
end

function world.update()
	game.goal = 0
	for i, v in pairs(entities) do
		if (v.x^2+v.y^2)^0.5 + v.r > world.radius then
			kill(i)
		end
		if v.breed ~= "gravitator" then
			world.matter = world.matter + v.m
			game.goal = math.max(v.m, game.goal)
		end
	end
	
	if game.status == nil then
		if entities[player.i] == nil then
			game.status = "lost"
		elseif entities[player.i].m >= game.goal then
			game.status = "won"
			map = "gravity II"
		end
	end
end

local units = 600
local exponent = 4

-- 1) Die Menge der Masse wird angegeben.
-- 2) Die Menge der gewnschten Lebensformen wird angegeben
-- 3) Aus der Menge der Lebensformen wird die Anzahl der Abstufungen berechnet (units^0.5)

breeds = {} -- load the different types of breeds here
love.filesystem.include("breeds/player/definition.lua")
love.filesystem.include("breeds/default/definition.lua")
love.filesystem.include("breeds/gravitator/definition.lua")

-- create gravitator
local gravitatormass = world.matter/2
create_entity("gravitator", gravitatormass ,0,0)

-- create player
local playerdistance = radius(gravitatormass) + world.radius*0.4
create_entity("player", world.matter/units*0.5 ,0, playerdistance, 1, 0)

-- create 100% of the units in region A orbiting the gravitator counter-clockwise
local i = 0
while i < units do
	i = i+1
	local matter = (math.random(1, 70)/10)^exponent/(10^(exponent-1))*world.matter/units
	local distance = math.random(10*(radius(gravitatormass) + radius(matter) + world.radius*0.1), 10*(world.radius*0.9 - radius(matter)))/10
	local angular = math.random(0, 360)
	
	local x = distance * math.cos(math.rad(angular))
	local y = distance * math.sin(math.rad(angular))
	
	local distance = (x^2+y^2)^0.5
	
	local create = true
	for j, v in pairs(entities) do
		local d = ((x-v.x)^2+(y-v.y)^2)^0.5
		local r = radius(matter)+v.r
		if d < r*1.1 then
			i = i -1
			create = false
			break
		end
	end
	
	if create == true then
		create_entity("default", matter, x, y, y/distance, -x/distance)
	end
end
